import json
import logging
import os
import subprocess
import warnings
from pathlib import Path

from .backend import settings
from .menu.screen import show_menu
from .misc.console import ConsoleStyle
from .utils import CustomUserWarning

# do not remove - hides the libraries' debug messages
logging.getLogger("urllib3").setLevel(logging.WARNING)
logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("tika").setLevel(logging.ERROR)
logging.getLogger("httpx").setLevel(logging.ERROR)
logging.getLogger("httpcore").setLevel(logging.ERROR)
logging.getLogger("huggingface_hub").setLevel(logging.ERROR)
logging.getLogger("huggingface").setLevel(logging.ERROR)
logging.getLogger('pydub').setLevel(logging.ERROR)

warnings.simplefilter("ignore")


# set the environment variable for the transformers library
os.environ['TOKENIZERS_PARALLELISM'] = "false"


SYMAI_VERSION = "0.7.5"
__version__   = SYMAI_VERSION
__root_dir__  = settings.HOME_PATH / '.symai'


def _start_symai():
    global _symai_config_
    global _symsh_config_
    global _symserver_config_

    # CREATE THE SYMAI FOLDER IF IT DOES NOT EXIST YET
    # *==============================================================================================================*
    if not os.path.exists(__root_dir__):
        os.makedirs(__root_dir__)

    # CREATE THE SHELL CONFIGURATION FILE IF IT DOES NOT EXIST YET
    # *==============================================================================================================*
    _symsh_config_path_ = __root_dir__ / 'symsh.config.json'
    if not os.path.exists(_symsh_config_path_):
        with open(_symsh_config_path_, "w") as f:
            json.dump({
                "colors": {
                    "completion-menu.completion.current": "bg:#323232 #212121",
                    "completion-menu.completion":         "bg:#800080 #212121",
                    "scrollbar.background":               "bg:#222222",
                    "scrollbar.button":                   "bg:#776677",
                    "history-completion":                 "bg:#212121 #f5f5f5",
                    "path-completion":                    "bg:#800080 #f5f5f5",
                    "file-completion":                    "bg:#9040b2 #f5f5f5",
                    "history-completion-selected":        "bg:#efefef #b3d7ff",
                    "path-completion-selected":           "bg:#efefef #b3d7ff",
                    "file-completion-selected":           "bg:#efefef #b3d7ff"
                },
                "map-nt-cmd":                             True,
                "show-splash-screen":                     True,
                "plugin_prefix":                          None
            }, f, indent=4)

    # CREATE A SERVER CONFIGURATION FILE IF IT DOES NOT EXIST YET
    # *==============================================================================================================*
    _symserver_config_path_ = __root_dir__ / 'symserver.config.json'
    if not os.path.exists(_symserver_config_path_):
        with open(_symserver_config_path_, "w") as f:
            json.dump({}, f, indent=4)

    # CHECK IF THE USER HAS A CONFIGURATION FILE IN THE CURRENT WORKING DIRECTORY (DEBUGGING MODE)
    # *==============================================================================================================*
    if os.path.exists(Path.cwd() / 'symai.config.json'):
        logging.debug('Using the configuration file in the current working directory.')
        _symai_config_path_ = Path.cwd() / 'symai.config.json'

    else:
        # CREATE THE CONFIGURATION FILE IF IT DOES NOT EXIST YET WITH THE DEFAULT VALUES
        # *==========================================================================================================*
        _symai_config_path_ = __root_dir__ / 'symai.config.json'

        if not os.path.exists(_symai_config_path_):
            setup_wizard(_symai_config_path_, show_wizard=False)
            CustomUserWarning('No configuration file found. A new configuration file has been created in your home directory. Please run the setup wizard in your console using the `symwzd` command or manually set your `.symai/symai.config.json` config situated in your home directory or set the environment variables for the respective engines.')

        # LOAD THE CONFIGURATION FILE
        # *==========================================================================================================*
        with open(_symai_config_path_, 'r', encoding="utf-8") as f:
            _symai_config_ = json.load(f)
        _tmp_symai_config_ = _symai_config_.copy()

        # MIGRATE THE ENVIRONMENT VARIABLES
        # *==========================================================================================================*
        if 'SPEECH_ENGINE_MODEL' in _symai_config_:
            _symai_config_['SPEECH_TO_TEXT_ENGINE_MODEL']     = _symai_config_['SPEECH_ENGINE_MODEL']
            _tmp_symai_config_['SPEECH_TO_TEXT_ENGINE_MODEL'] = _symai_config_['SPEECH_ENGINE_MODEL']
            del _symai_config_['SPEECH_ENGINE_MODEL']
            del _tmp_symai_config_['SPEECH_ENGINE_MODEL']
            # create missing environment variable
            _symai_config_['TEXT_TO_SPEECH_ENGINE_MODEL']     = "tts-1"
            # save the updated configuration file
            with open(_symai_config_path_, 'w') as f:
                json.dump(_symai_config_, f, indent=4)

        if 'COLLECTION_URI' not in _symai_config_:
            print('Migrating the configuration file to the latest version.')
            _symai_config_['COLLECTION_URI']     = "mongodb+srv://User:vt3epocXitd6WlQ6@extensityai.c1ajxxy.mongodb.net/?retryWrites=true&w=majority"
            _symai_config_['COLLECTION_DB']      = "ExtensityAI"
            _symai_config_['COLLECTION_STORAGE'] = "SymbolicAI"
            _symai_config_['SUPPORT_COMMUNITY']  = False
            with ConsoleStyle('info') as console:
                msg = 'Currently you are sharing your user experience with us by uploading the data to our research server, and thereby helping us improve future models and the overall SymbolicAI experience. We thank you very much for supporting the research community! If you wish to disable the data collection option go to your .symai config situated in your home directory or set the environment variable `SUPPORT_COMMUNITY` to `False`.'
                console.print(msg)
            # save the updated configuration file
            with open(_symai_config_path_, 'w') as f:
                json.dump(_symai_config_, f, indent=4)

        # VERIFY IF THE CONFIGURATION FILE HAS CHANGED AND UPDATE IT
        # *==========================================================================================================*
        _updated_key_ = {k: not _tmp_symai_config_[k] == _symai_config_[k] for k in _tmp_symai_config_.keys()}
        _has_changed_ = any(_updated_key_.values())

        if _has_changed_:
            # update the symai.config.json file
            with open(_symai_config_path_, 'w') as f:
                json.dump(_symai_config_, f, indent=4)

        # POST-MIGRATION CHECKS
        # *==============================================================================================================*
        # CHECK IF THE USER HAS A TEXT TO SPEECH ENGINE API KEY
        if 'TEXT_TO_SPEECH_ENGINE_API_KEY' not in _symai_config_:
            _symai_config_['TEXT_TO_SPEECH_ENGINE_API_KEY']     = _symai_config_['NEUROSYMBOLIC_ENGINE_API_KEY'] if 'NEUROSYMBOLIC_ENGINE_API_KEY' in _symai_config_ else ''
            _tmp_symai_config_['TEXT_TO_SPEECH_ENGINE_API_KEY'] = _symai_config_['NEUROSYMBOLIC_ENGINE_API_KEY'] if 'NEUROSYMBOLIC_ENGINE_API_KEY' in _symai_config_ else ''
            # save the updated configuration file
            with open(_symai_config_path_, 'w') as f:
                json.dump(_symai_config_, f, indent=4)

    # LOAD THE CONFIGURATION FILE
    # *==============================================================================================================*
    with open(_symai_config_path_, 'r', encoding="utf-8") as f:
        _symai_config_ = json.load(f)

    # LOAD THE SHELL CONFIGURATION FILE
    # *==============================================================================================================*
    with open(_symsh_config_path_, 'r', encoding="utf-8") as f:
        _symsh_config_ = json.load(f)

    # LOAD THE SERVER CONFIGURATION FILE
    # *==============================================================================================================*
    with open(_symserver_config_path_, 'r', encoding="utf-8") as f:
        _symserver_config_ = json.load(f)

    # MIGRATE THE SHELL SPLASH SCREEN CONFIGURATION
    # *==============================================================================================================*
    if 'show-splash-screen' not in _symsh_config_:
        _symsh_config_['show-splash-screen'] = True
        with open(_symsh_config_path_, 'w') as f:
            json.dump(_symsh_config_, f, indent=4)

    # CHECK IF THE USER HAS A NEUROSYMBOLIC API KEY
    # *==============================================================================================================*
    if not (
        _symai_config_['NEUROSYMBOLIC_ENGINE_MODEL'].lower().startswith('llama') or \
        _symai_config_['NEUROSYMBOLIC_ENGINE_MODEL'].lower().startswith('huggingface')) \
        and \
            (
            _symai_config_['NEUROSYMBOLIC_ENGINE_API_KEY'] is None or \
            len(_symai_config_['NEUROSYMBOLIC_ENGINE_API_KEY']) == 0):

        CustomUserWarning('The mandatory neuro-symbolic engine is not initialized. Please set the NEUROSYMBOLIC_ENGINE_MODEL and NEUROSYMBOLIC_ENGINE_API_KEY environment variables.')

    settings.SYMAI_CONFIG = _symai_config_
    settings.SYMSH_CONFIG = _symsh_config_
    settings.SYMSERVER_CONFIG = _symserver_config_


def run_setup_wizard(file_path = __root_dir__ / 'symai.config.json'):
    setup_wizard(file_path)


def run_server():
    _symserver_config_ = {}
    if settings.SYMAI_CONFIG.get("NEUROSYMBOLIC_ENGINE_MODEL").startswith("llama"):
        from .server.llama_cpp_server import llama_cpp_server

        command, args = llama_cpp_server()
        _symserver_config_.update(zip(args[::2], args[1::2]))
        _symserver_config_['online'] = True

        with open(__root_dir__ / "symserver.config.json", "w") as f:
            json.dump(_symserver_config_, f, indent=4)

        try:
            subprocess.run(command, check=True)
        except KeyboardInterrupt:
            print("Server stopped!")
        except Exception as e:
            print(f"Error running server: {e}")
        finally:
            with open(__root_dir__ / "symserver.config.json", "w") as f:
                json.dump({'online': False}, f, indent=4)

    elif settings.SYMAI_CONFIG.get("NEUROSYMBOLIC_ENGINE_MODEL").startswith("huggingface"):
        from .server.huggingface_server import huggingface_server

        command, args = huggingface_server()
        _symserver_config_.update(vars(args))
        _symserver_config_['online'] = True

        with open(__root_dir__ / "symserver.config.json", "w") as f:
            json.dump(_symserver_config_, f, indent=4)

        try:
            command(host=args.host, port=args.port)
        except KeyboardInterrupt:
            print("Server stopped!")
        except Exception as e:
            print(f"Error running server: {e}")
        finally:
            with open(__root_dir__ / "symserver.config.json", "w") as f:
                json.dump({'online': False}, f, indent=4)
    else:
        raise CustomUserWarning("You're trying to run a local server without a valid neuro-symbolic engine model. Please set a valid model in your configuration file. Current available options are 'llamacpp' and 'huggingface'.")


def setup_wizard(_symai_config_path_, show_wizard=True):
    _user_config_                   = show_menu(show_wizard=show_wizard)
    _nesy_engine_api_key            = _user_config_['nesy_engine_api_key']
    _nesy_engine_model              = _user_config_['nesy_engine_model']
    _symbolic_engine_api_key        = _user_config_['symbolic_engine_api_key']
    _symbolic_engine_model          = _user_config_['symbolic_engine_model']
    _embedding_engine_api_key       = _user_config_['embedding_engine_api_key']
    _embedding_model                = _user_config_['embedding_model']
    _drawing_engine_api_key         = _user_config_['drawing_engine_api_key']
    _drawing_engine_model           = _user_config_['drawing_engine_model']
    _vision_engine_model            = _user_config_['vision_engine_model']
    _search_engine_api_key          = _user_config_['search_engine_api_key']
    _search_engine_model            = _user_config_['search_engine_model']
    _ocr_engine_api_key             = _user_config_['ocr_engine_api_key']
    _speech_to_text_engine_model    = _user_config_['speech_to_text_engine_model']
    _text_to_speech_engine_api_key  = _user_config_['text_to_speech_engine_api_key']
    _text_to_speech_engine_model    = _user_config_['text_to_speech_engine_model']
    _text_to_speech_engine_voice    = _user_config_['text_to_speech_engine_voice']
    _indexing_engine_api_key        = _user_config_['indexing_engine_api_key']
    _indexing_engine_environment    = _user_config_['indexing_engine_environment']
    _caption_engine_environment     = _user_config_['caption_engine_environment']
    _support_comminity              = _user_config_['support_community']

    with open(_symai_config_path_, 'w') as f:
        json.dump({
            "NEUROSYMBOLIC_ENGINE_API_KEY":   _nesy_engine_api_key,
            "NEUROSYMBOLIC_ENGINE_MODEL":     _nesy_engine_model,
            "SYMBOLIC_ENGINE_API_KEY":        _symbolic_engine_api_key,
            "SYMBOLIC_ENGINE":                _symbolic_engine_model,
            "EMBEDDING_ENGINE_API_KEY":       _embedding_engine_api_key,
            "EMBEDDING_ENGINE_MODEL":         _embedding_model,
            "DRAWING_ENGINE_API_KEY":         _drawing_engine_api_key,
            "DRAWING_ENGINE_MODEL":           _drawing_engine_model,
            "VISION_ENGINE_MODEL":            _vision_engine_model,
            "SEARCH_ENGINE_API_KEY":          _search_engine_api_key,
            "SEARCH_ENGINE_MODEL":            _search_engine_model,
            "OCR_ENGINE_API_KEY":             _ocr_engine_api_key,
            "SPEECH_TO_TEXT_ENGINE_MODEL":    _speech_to_text_engine_model,
            "TEXT_TO_SPEECH_ENGINE_API_KEY":  _text_to_speech_engine_api_key,
            "TEXT_TO_SPEECH_ENGINE_MODEL":    _text_to_speech_engine_model,
            "TEXT_TO_SPEECH_ENGINE_VOICE":    _text_to_speech_engine_voice,
            "INDEXING_ENGINE_API_KEY":        _indexing_engine_api_key,
            "INDEXING_ENGINE_ENVIRONMENT":    _indexing_engine_environment,
            "CAPTION_ENGINE_MODEL":           _caption_engine_environment,
            "COLLECTION_URI":                 "mongodb+srv://User:vt3epocXitd6WlQ6@extensityai.c1ajxxy.mongodb.net/?retryWrites=true&w=majority",
            "COLLECTION_DB":                  "ExtensityAI",
            "COLLECTION_STORAGE":             "SymbolicAI",
            "SUPPORT_COMMUNITY":              _support_comminity
        }, f, indent=4)


_start_symai()


from .backend.base import Engine
from .components import Function, PrimitiveDisabler
from .core import few_shot, zero_shot
from .extended import Conversation
from .functional import EngineRepository
from .imports import Import
from .interfaces import Interface
from .post_processors import PostProcessor
from .pre_processors import PreProcessor
from .prompts import Prompt, PromptRegistry, PromptLanguage
from .shell import Shell
from .strategy import Strategy
from .symbol import Call, Expression, GlobalSymbolPrimitive, Metadata, Symbol
